#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/cacheflush.h>


#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ide.h>

#include "Types.h"

#include "agentlite.h"
#include "support_fpga.h"
#include "support_fpga_hw.h"

#if 0
#define ADC_MEM_CHECK_REG_FAILS		0xBB
#define ADC_REQ_MEM_REG_FAILS		0xCC
#define ADC_MEM_REMAP_FAILS			0x0A
#define ADC_REG_CHR_DEV_FAILS 		0xAA
#define SET_ADC_OPTION				4
#endif

#define MAX_RETRIES 3
#define INVALID_EXCHANGE_BUS_WAIT_PERIOD  HZ/(10 * MAX_RETRIES)



#define SUPPORT_FPGA_MEM_REMAP_FAILS 0xAB 
#define SUPPORT_FPGA_MEM_CHECK_REG_FAILS 0xAC 
#define SUPPORT_FPGA_REQ_MEM_REG_FAILS 0xAD 
#define SUPPORT_FPGA_REG_CHR_DEV_FAILS 		0xAE



#define SIMULATION 1


#ifdef SIMULATION
static unsigned short simu_data[256] = {0};
#endif



/* ------------------------------------------------------------------------- *
 *  Function prototypes
 * ------------------------------------------------------------------------- */



static int support_fpga_int(void );
static void support_fpga_exit(void );
static int support_fpga_module_open(struct inode *inode, struct file *file);
static int support_fpga_module_close(struct inode *inode, struct file *file);
static long support_fpga_module_ioctl(struct file *file, uint cmd, ulong arg);
extern int write_init_st(int option,int value);



/* ------------------------------------------------------------------------- 
 *  Static variables 
 * ------------------------------------------------------------------------- */
static struct file_operations support_fpga_module_ops;
static void *support_fpga_parallel_bus_remap_addr ;
static short init_support_fpga_status=0;



/* SUPPORT FPGA Driver Private data */
typedef struct {

	int (*read_parallel_bus_reg)  ( void * support_fpga_base_addr, uint16_t * support_fpga_value, uint32_t reg_offset );
	int (*write_parallel_bus_reg)  ( void * support_fpga_base_addr, uint16_t support_fpga_value, uint32_t reg_offset );
	int (*request_i2c_shared_bus) ( void * support_fpga_base_addr, uint8_t segnum);
	int (*granted_i2c_shared_bus) ( void * support_fpga_base_addr, uint8_t segnum);
	int (*release_i2c_shared_bus) ( void * support_fpga_base_addr, uint8_t segnum);
	
   }  __attribute__((packed)) support_fpga_drv_t; 







/* ------------------------------------------------------------------------- 
 *  Functions
 * ------------------------------------------------------------------------- */

/*** iowrite ***/
void support_fpga_write_reg(unsigned long addr, unsigned long val) {
    *((volatile uint32_t*)addr) = val;
}



/*** ioread ***/
unsigned long  support_fpga_read_reg(unsigned long addr ) {
    return ((*(volatile uint32_t*)addr));
}




static int
support_fpga_int(void )
{
	int err,ret=0;
	uint32_t support_fpga_base;

	

	memset(&support_fpga_module_ops,0,sizeof(support_fpga_module_ops));
	support_fpga_module_ops.open=support_fpga_module_open;
	support_fpga_module_ops.release=support_fpga_module_close;
	support_fpga_module_ops.unlocked_ioctl= support_fpga_module_ioctl;
	support_fpga_module_ops.read=NULL;
	support_fpga_module_ops.write=NULL;
	support_fpga_module_ops.owner=THIS_MODULE;

    /* ----  Register the character device ------------------- */
	if ((err = register_chrdev(SUPPORT_FPGA_MAJOR_NUM,DRIVER_NAME,&support_fpga_module_ops)) < 0) {
		printk (KERN_ERR "support_fpga.c: SUPPORT FPGA register_chrdrv failed. err = %d\n", err);
		init_support_fpga_status=SUPPORT_FPGA_REG_CHR_DEV_FAILS;
		return -ENODEV;
	}

	/* Remap registers */
	 support_fpga_base= SUPPORT_FPGA_BASE_OFFSET;
	if ((support_fpga_parallel_bus_remap_addr = ioremap_nocache(support_fpga_base, SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE)) == NULL) {
		printk(KERN_ERR "failed to map SUPPORT FPGA  IO space %08x-%08x to memory",support_fpga_base, support_fpga_base + SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE - 1);
		init_support_fpga_status=SUPPORT_FPGA_MEM_REMAP_FAILS;
		ret = -EIO;
		support_fpga_exit();
		return ret;
	}

	/* check Support FPGA memory region */
	if ((err = check_mem_region(support_fpga_base, SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE)) < 0)  {
		printk(KERN_ERR "Support FPGA IO space %08x-%08x already in use (err %d)",support_fpga_base, support_fpga_base + SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE - 1, err);
		init_support_fpga_status=SUPPORT_FPGA_MEM_CHECK_REG_FAILS;
		ret = -EBUSY;
		support_fpga_exit();
		return ret;
	}

	/* request for the memory region */	
	if (request_mem_region(support_fpga_base,SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE,DRIVER_NAME)  == NULL) { 
		printk(KERN_ERR "IO space %08x-%08x adc request mem region failed (err %d)",support_fpga_base, support_fpga_base + SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE - 1, err);
		init_support_fpga_status=SUPPORT_FPGA_REQ_MEM_REG_FAILS;
		ret = -ENODEV;
		support_fpga_exit();
		return ret;
	}


   
	//write_init_st(SET_ADC_OPTION,init_adc_status);	
	printk(KERN_ALERT "Support FPGA Module was  sucessully loaded \n");
	return ret;
    
}


void 
support_fpga_exit(void ) {
	uint32_t support_fpga_base;
	

	printk (KERN_ALERT "Unloading Support FPGA driver ..\n");
	support_fpga_base = SUPPORT_FPGA_BASE_OFFSET;
	
	if(init_support_fpga_status == 0) {
		init_support_fpga_status = 0xff;
	}
	

	/* release the memory occupied */  
	release_mem_region(support_fpga_base,SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE);

	/* Unmap registers */
	iounmap( support_fpga_parallel_bus_remap_addr); 
	
	
	/* ----  UnRegister the character device ------------------- */
	unregister_chrdev(SUPPORT_FPGA_MAJOR_NUM,DRIVER_NAME);

	return ;

}



int 
support_fpga_module_open(struct inode *inode, struct file *file) {

   support_fpga_drv_t *support_fpga_drv  = (support_fpga_drv_t*) kmalloc ( sizeof(support_fpga_drv_t), GFP_KERNEL );
	memset(support_fpga_drv, 0, sizeof(support_fpga_drv_t));


	/* Initialize the private data */	
	support_fpga_drv->read_parallel_bus_reg = support_fpga_read_parallel_bus_reg;
	support_fpga_drv->write_parallel_bus_reg = support_fpga_write_parallel_bus_reg;
	support_fpga_drv->request_i2c_shared_bus = request_i2c_shared_bus;
	support_fpga_drv->granted_i2c_shared_bus = granted_i2c_shared_bus;
	support_fpga_drv->release_i2c_shared_bus = release_i2c_shared_bus;

	file->private_data = support_fpga_drv;



	return 0;
}


int 
support_fpga_module_close(struct inode *inode, struct file *file) {
	support_fpga_drv_t *support_fpga_drv;
	support_fpga_drv = file->private_data;
	file->private_data = NULL;
	kfree(support_fpga_drv);


	return 0;
}


long 
support_fpga_module_ioctl(struct file *file, uint cmd, ulong arg) {
	support_fpga_drv_t *support_fpga_drv;
	reg_value_t  reg_value;
 

	support_fpga_drv =file->private_data; 

	switch (cmd) {	
		case READ_REG :
			copy_from_user ((void*)&reg_value, (void *)arg,sizeof (reg_value_t));
	#ifdef SIMULATION
				reg_value.reg_offset_addr = reg_value.reg_offset_addr & 0xFF; // simulation data only has 256 bytes.
				reg_value.reg_value = simu_data [reg_value.reg_offset_addr];

	#else 
				if (support_fpga_drv->read_parallel_bus_reg)
				{
					if (0 != support_fpga_drv->read_parallel_bus_reg(support_fpga_parallel_bus_remap_addr,&reg_value.reg_value,reg_value.reg_offset_addr ))
					return -1;
				}
			
	#endif
			copy_to_user ((void*)arg, (void*)&reg_value, sizeof (reg_value_t));
			break;

		case WRITE_REG :
			copy_from_user ((void*)&reg_value, (void *)arg,sizeof (reg_value_t));
	#ifdef SIMULATION
			reg_value.reg_offset_addr = reg_value.reg_offset_addr & 0xFF; // simulation data only has 256 bytes.
			simu_data [reg_value.reg_offset_addr] = reg_value.reg_value;
	#else 

			if (support_fpga_drv->write_parallel_bus_reg)
			{
				if (0 != support_fpga_drv->write_parallel_bus_reg(support_fpga_parallel_bus_remap_addr,reg_value.reg_value,reg_value.reg_offset_addr ))
				return -1;
			}
	#endif
			break;

		case I2C_REQUEST_SHARED_BUS:
			copy_from_user ((void*)&reg_value, (void *)arg,sizeof (reg_value_t));

			if (support_fpga_drv->request_i2c_shared_bus) {
				if ( 0!= support_fpga_drv->request_i2c_shared_bus(support_fpga_parallel_bus_remap_addr,reg_value.segnum))
					return -1;
			}
			//copy_to_user ((void*)arg, (void*)&reg_value, sizeof (reg_value_t));
			break;

		case I2C_GRANT_SHRARED_BUS:
			copy_from_user ((void*)&reg_value, (void *)arg,sizeof (reg_value_t));

			if (support_fpga_drv->granted_i2c_shared_bus) {
				if ( 0!= support_fpga_drv->granted_i2c_shared_bus(support_fpga_parallel_bus_remap_addr,reg_value.segnum))
					return -1;
			}
			//copy_to_user ((void*)arg, (void*)&reg_value, sizeof (reg_value_t));
			break;

		case I2C_RELEASE_SHARED_BUS:
			
			copy_from_user ((void*)&reg_value, (void *)arg,sizeof (reg_value_t));
			
			if (support_fpga_drv->release_i2c_shared_bus) {
				if ( 0!= support_fpga_drv->release_i2c_shared_bus(support_fpga_parallel_bus_remap_addr,reg_value.segnum))
					return -1;
			}
			//copy_to_user ((void*)arg, (void*)&reg_value, sizeof (reg_value_t));
			break;
		default :
			printk (KERN_ALERT "Invalid ADC driver ioctl\n");
			return -EINVAL;
		}
	
	return 0;
}


module_init(support_fpga_int);
module_exit(support_fpga_exit);

int request_i2c_shared_bus(void * support_fpga_base_addr, uint8_t segnum){
unsigned short val_mlbr;
	unsigned short val_mler;
	dev_status_t retval=DEV_STATUS_SUCCESS;
	
	/***** Read local bus register *****/
	support_fpga_read_parallel_bus_reg(support_fpga_base_addr, &val_mlbr, MLBR_REG);
	/******Read local error register ****/	
	support_fpga_read_parallel_bus_reg(support_fpga_base_addr, &val_mler, MLER_REG); 
	
	/**** If (local bus request is already asserted(MLBR[4]set) ***/
	if (val_mlbr & MLBR_LOC_BUS_REQ) {
		printk("Local bus request is already asserted\n");
		retval=DEV_STATUS_INACCESSIBLE;
	}
	
	/**** If (local bus request is already active(MLBR[5]set) ***/
	if (val_mlbr & MLBR_LOC_BUS_ACT) {
	
		printk("Local bus request is already active\n");
		retval=DEV_STATUS_INACCESSIBLE;
	} 

	if (((val_mlbr & MLBR_LOC_BUS_ACT) == 0) && ((val_mler & MLER_LOC_REL_SHORT_TMO) == 0) 
			&& ((val_mler & MLER_LOC_REL_LONG_TMO) == 0)) {
		/******* Set request bus number (segment number) *******/
		support_fpga_write_parallel_bus_reg (support_fpga_base_addr,segnum | val_mlbr,MLBR_REG); 

		/******* Set the access type to READ/WRITE *****/
		support_fpga_write_parallel_bus_reg (support_fpga_base_addr,MLBR_LOC_ACC_TYPE | val_mlbr,MLBR_REG); 
	}

	/******* request bus ************/
	support_fpga_write_parallel_bus_reg (support_fpga_base_addr,MLBR_LOC_BUS_REQ | val_mlbr,MLBR_REG); 
	
	return retval;
}

int granted_i2c_shared_bus(void * support_fpga_base_addr, uint8_t segnum){

	int retry = 0;
	unsigned short val_mlbr,tmp;
	unsigned short val_mler,val_mrer,val_mcsr;
	unsigned long delay=35000;
	unsigned long invalid_bus_waiting=1000 * INVALID_EXCHANGE_BUS_WAIT_PERIOD ;
	dev_status_t retval=DEV_STATUS_SUCCESS;
	
	/***** Read Local Bus  Register *****/
	support_fpga_read_parallel_bus_reg(support_fpga_base_addr, &val_mlbr,  MLBR_REG);

	/*********** while (Bus is not active (! MLBR[5]) and retry < Max Retries) ****/
	while (!(val_mlbr & MLBR_LOC_BUS_ACT) && (retry < MAX_RETRIES)) {


		/**** Read local errors(MLER) and remote errors(MRER)****/
		support_fpga_read_parallel_bus_reg(support_fpga_base_addr, &val_mler, MLER_REG &val_mler);
		support_fpga_read_parallel_bus_reg(support_fpga_base_addr, &val_mrer, MRER_REG); 

		if ((val_mler != 0) || (val_mrer != 0))
			printk("Local Error or Remote Error\n");

		/***** If ( Remote Grant Timeout (MLER[2]) ****/
		if (val_mler & MLER_RM_GRT_TMO)
			retry ++;

		/****Else if (Remote Release Timeout (MLER[3]) ***/
		else if (val_mler & MLER_RM_REL_TMO) {
				
			udelay (1000);
			retry ++;
		}
		/** Else if (Local Invalid Serial Exchange Bus[MLER[5]) or Remote Invalid Serial Exchange Bus(MRER[5]) **/
		else if ((val_mler & MLER_LOC_INV_SER_EXCH_BUS) || (val_mrer & MRER_RM_INV_SER_EXCH_BUS)) {
			udelay (invalid_bus_waiting);
			retry ++;
		}
		/** Else if (Local Release Short Timeout(MLER[0]) or Local Release Long Timeout(MLER[1])**/
		else if ((val_mler & MLER_LOC_REL_SHORT_TMO) || (val_mler & MLER_LOC_REL_LONG_TMO) ) {
			printk("Bus is inaccessible due to FPGA error\n");
			retval=DEV_STATUS_INACCESSIBLE;
		}
	
	}

	/***** If (Retry == max retries)enable remote override ****/
	if (retry == MAX_RETRIES) {
		support_fpga_read_parallel_bus_reg(support_fpga_base_addr, &val_mcsr,SUPFPGA_MCSR_REG); 
			
		tmp = ~MCSR_RM_CTRL_OVR_ENABLE & val_mcsr; // clear MCSR bit 5 and bit 4;

		/****** Enable Remote Override( Write override sequence (MCSR[5:4]) *****/
		while (delay-- > 0) { //hopeful to generate 2ms interval
			support_fpga_write_parallel_bus_reg (support_fpga_base_addr,0x01 << 4 | tmp,MCSR_REG);
			support_fpga_write_parallel_bus_reg (support_fpga_base_addr,0x03 << 4 | tmp,MCSR_REG);
			support_fpga_write_parallel_bus_reg (support_fpga_base_addr,0x02 << 4 | tmp,MCSR_REG);
			support_fpga_write_parallel_bus_reg (support_fpga_base_addr,0x00 << 4 | tmp,MCSR_REG);
			
		}
		/**** Read Control and Status Register(MCSR) and Local Bus Register(MLBR)****/
		support_fpga_read_parallel_bus_reg(support_fpga_base_addr,&val_mcsr, MCSR_REG);
		support_fpga_read_parallel_bus_reg(support_fpga_base_addr,&val_mlbr, MLBR_REG); 

		/*****If remote Override is set (MCSR[1]) ****/
		if (val_mcsr & MCSR_RM_CTRL_OVR_STATE) {
			/**** If (Bus is not active (!MLBR[5]) ****/
			if (!(val_mlbr & MLBR_LOC_BUS_ACT)) {
				printk("Bus is inaccessible due to FPGA error\n");
				retval=DEV_STATUS_INACCESSIBLE;
			}
				
		}

	}
	return retval;

}

int 
release_i2c_shared_bus(void * support_fpga_base_addr, uint8_t segnum) {
	
	unsigned short val;
	dev_status_t retval=DEV_STATUS_SUCCESS;
	
	#if 0
	Clear local error register and remote error register;
	Set Release bus number (Set MLBR[2:0] = bus);
	Release Bus Request (Clear MLBR[4]);
	#endif

	/***** Read local bus register *****/
	support_fpga_read_parallel_bus_reg(support_fpga_base_addr,&val, MLBR_REG);
	/******* Set request bus number (segment number) *******/
	support_fpga_write_parallel_bus_reg (support_fpga_base_addr,segnum||val,MLBR_REG);
	
	/***** Release Bus Request (Clear MLBR[4]) ***/
	support_fpga_write_parallel_bus_reg (support_fpga_base_addr,~MLBR_LOC_BUS_REQ & val,MLBR_REG);
	return retval;
}


int
support_fpga_read_parallel_bus_reg( void * support_fpga_base_addr,  uint16_t*support_fpga_value, uint32_t reg_off_addr) {
	volatile uint32_t reg_buff =0;
		

	
	reg_buff = support_fpga_read_reg(( *(unsigned long *) support_fpga_base_addr) + reg_off_addr );
	  
	*support_fpga_value = (uint16_t)reg_buff;

	return 0;
}

int
support_fpga_write_parallel_bus_reg(void *support_fpga_base_addr,  uint16_t support_fpga_value, uint32_t reg_off_addr) {


	
	support_fpga_write_reg( *(unsigned long *)support_fpga_base_addr + reg_off_addr,  (uint32_t)support_fpga_value ); 

	return 0;

}






MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Support FPGA Driver");


